@@ -21,7 +21,7 @@ import time
 from os.path import dirname, basename
 
 from typing import (AbstractSet, Dict, Iterable, Iterator, List,
-                    NamedTuple, Optional, Set, Tuple, Union)
+                    NamedTuple, Optional, Set, Tuple, Union, Mapping)
 
 from mypy.types import Type
 from mypy.nodes import (MypyFile, Node, Import, ImportFrom, ImportAll,
@@ -35,6 +35,7 @@ from mypy import defaults
 from mypy import moduleinfo
 from mypy import util
 from mypy.fixup import fixup_module_pass_one, fixup_module_pass_two
+from mypy.options import Options
 from mypy.parse import parse
 from mypy.stats import dump_type_stats
 
@@ -43,32 +44,6 @@ from mypy.stats import dump_type_stats
 # until Python 3.4, __file__ is relative.
 __file__ = os.path.realpath(__file__)
 
-
-# Build targets (for selecting compiler passes)
-SEMANTIC_ANALYSIS = 0   # Semantic analysis only
-TYPE_CHECK = 1          # Type check
-
-
-# Build flags
-VERBOSE = 'verbose'              # More verbose messages (for troubleshooting)
-MODULE = 'module'                # Build module as a script
-PROGRAM_TEXT = 'program-text'    # Build command-line argument as a script
-TEST_BUILTINS = 'test-builtins'  # Use stub builtins to speed up tests
-DUMP_TYPE_STATS = 'dump-type-stats'
-DUMP_INFER_STATS = 'dump-infer-stats'
-SILENT_IMPORTS = 'silent-imports'  # Silence imports of .py files
-ALMOST_SILENT = 'almost-silent'  # If SILENT_IMPORTS: report silenced imports as errors
-INCREMENTAL = 'incremental'      # Incremental mode: use the cache
-FAST_PARSER = 'fast-parser'      # Use experimental fast parser
-# Disallow calling untyped functions from typed ones
-DISALLOW_UNTYPED_CALLS = 'disallow-untyped-calls'
-# Disallow defining untyped (or incompletely typed) functions
-DISALLOW_UNTYPED_DEFS = 'disallow-untyped-defs'
-# Type check unannotated functions
-CHECK_UNTYPED_DEFS = 'check-untyped-defs'
-# Also check typeshed for missing annotations
-WARN_INCOMPLETE_STUB = 'warn-incomplete-stub'
-
 PYTHON_EXTENSIONS = ['.pyi', '.py']
 
 
@@ -130,14 +105,9 @@ class BuildSourceSet:
 
 
 def build(sources: List[BuildSource],
-          target: int,
+          options: Options,
           alt_lib_path: str = None,
-          bin_dir: str = None,
-          pyversion: Tuple[int, int] = defaults.PYTHON3_VERSION,
-          custom_typing_module: str = None,
-          report_dirs: Dict[str, str] = None,
-          flags: List[str] = None,
-          python_path: bool = False) -> BuildResult:
+          bin_dir: str = None) -> BuildResult:
     """Analyze a program.
 
     A single call to build performs parsing, semantic analysis and optionally
@@ -147,30 +117,27 @@ def build(sources: List[BuildSource],
     otherwise raise CompileError.
 
     Args:
-      target: select passes to perform (a build target constant, e.g. C)
       sources: list of sources to build
+      options: build options
       alt_lib_path: an additional directory for looking up library modules
         (takes precedence over other directories)
       bin_dir: directory containing the mypy script, used for finding data
         directories; if omitted, use '.' as the data directory
-      pyversion: Python version (major, minor)
-      custom_typing_module: if not None, use this module id as an alias for typing
-      flags: list of build options (e.g. COMPILE_ONLY)
     """
-    report_dirs = report_dirs or {}
-    flags = flags or []
 
     data_dir = default_data_dir(bin_dir)
 
     find_module_clear_caches()
 
     # Determine the default module search path.
-    lib_path = default_lib_path(data_dir, pyversion, python_path)
+    lib_path = default_lib_path(data_dir, options.python_version)
 
-    if TEST_BUILTINS in flags:
+    if options.use_builtins_fixtures:
         # Use stub builtins (to speed up test cases and to make them easier to
-        # debug).
-        lib_path.insert(0, os.path.join(os.path.dirname(__file__), 'test', 'data', 'lib-stub'))
+        # debug).  This is a test-only feature, so assume our files are laid out
+        # as in the source tree.
+        root_dir = os.path.dirname(os.path.dirname(__file__))
+        lib_path.insert(0, os.path.join(root_dir, 'test-data', 'unit', 'lib-stub'))
     else:
         for source in sources:
             if source.path:
@@ -193,19 +160,18 @@ def build(sources: List[BuildSource],
     if alt_lib_path:
         lib_path.insert(0, alt_lib_path)
 
-    reports = Reports(data_dir, report_dirs)
+    reports = Reports(data_dir, options.report_dirs)
 
     source_set = BuildSourceSet(sources)
 
     # Construct a build manager object to hold state during the build.
     #
     # Ignore current directory prefix in error messages.
-    manager = BuildManager(data_dir, lib_path, target,
-                           pyversion=pyversion, flags=flags,
+    manager = BuildManager(data_dir, lib_path,
                            ignore_prefix=os.getcwd(),
-                           custom_typing_module=custom_typing_module,
                            source_set=source_set,
-                           reports=reports)
+                           reports=reports,
+                           options=options)
 
     try:
         dispatch(sources, manager)
@@ -219,7 +185,7 @@ def build(sources: List[BuildSource],
         reports.finish()
 
 
-def default_data_dir(bin_dir: str) -> str:
+def default_data_dir(bin_dir: Optional[str]) -> str:
     """Returns directory containing typeshed directory
 
     Args:
@@ -278,8 +244,7 @@ def mypy_path() -> List[str]:
     return path_env.split(os.pathsep)
 
 
-def default_lib_path(data_dir: str, pyversion: Tuple[int, int],
-        python_path: bool) -> List[str]:
+def default_lib_path(data_dir: str, pyversion: Tuple[int, int]) -> List[str]:
     """Return default standard library search paths."""
     # IDEA: Make this more portable.
     path = []  # type: List[str]
@@ -304,12 +269,6 @@ def default_lib_path(data_dir: str, pyversion: Tuple[int, int],
     if sys.platform != 'win32':
         path.append('/usr/local/lib/mypy')
 
-    # Contents of Python's sys.path go last, to prefer the stubs
-    # TODO: To more closely model what Python actually does, builtins should
-    #       go first, then sys.path, then anything in stdlib and third_party.
-    if python_path:
-        path.extend(sys.path)
-
     return path
 
 
@@ -322,13 +281,23 @@ CacheMeta = NamedTuple('CacheMeta',
                         ('data_mtime', float),  # mtime of data_json
                         ('data_json', str),  # path of <id>.data.json
                         ('suppressed', List[str]),  # dependencies that weren't imported
-                        ('flags', Optional[List[str]]),  # build flags
+                        ('options', Optional[Dict[str, bool]]),  # build options
+                        ('dep_prios', List[int]),
                         ])
-# NOTE: dependencies + suppressed == all unreachable imports;
+# NOTE: dependencies + suppressed == all reachable imports;
 # suppressed contains those reachable imports that were prevented by
 # --silent-imports or simply not found.
 
 
+# Priorities used for imports.  (Here, top-level includes inside a class.)
+# These are used to determine a more predictable order in which the
+# nodes in an import cycle are processed.
+PRI_HIGH = 5  # top-level "from X import blah"
+PRI_MED = 10  # top-level "import X"
+PRI_LOW = 20  # either form inside a function
+PRI_ALL = 99  # include all priorities
+
+
 class BuildManager:
     """This class holds shared state for building a mypy program.
 
@@ -338,7 +307,6 @@ class BuildManager:
 
     Attributes:
       data_dir:        Mypy data directory (contains stubs)
-      target:          Build target; selects which passes to perform
       lib_path:        Library path for looking up modules
       modules:         Mapping of module ID to MypyFile (shared by the passes)
       semantic_analyzer:
@@ -347,53 +315,40 @@ class BuildManager:
                        Semantic analyzer, pass 3
       type_checker:    Type checker
       errors:          Used for reporting all errors
-      pyversion:       Python version (major, minor)
-      flags:           Build options
+      options:         Build options
       missing_modules: Set of modules that could not be imported encountered so far
+      stale_modules:   Set of modules that needed to be rechecked
     """
 
     def __init__(self, data_dir: str,
                  lib_path: List[str],
-                 target: int,
-                 pyversion: Tuple[int, int],
-                 flags: List[str],
                  ignore_prefix: str,
-                 custom_typing_module: str,
                  source_set: BuildSourceSet,
-                 reports: Reports) -> None:
+                 reports: Reports,
+                 options: Options) -> None:
         self.start_time = time.time()
         self.data_dir = data_dir
         self.errors = Errors()
         self.errors.set_ignore_prefix(ignore_prefix)
         self.lib_path = tuple(lib_path)
-        self.target = target
-        self.pyversion = pyversion
-        self.flags = flags
-        self.custom_typing_module = custom_typing_module
         self.source_set = source_set
         self.reports = reports
-        check_untyped_defs = CHECK_UNTYPED_DEFS in self.flags
-        self.semantic_analyzer = SemanticAnalyzer(lib_path, self.errors,
-                                                  pyversion=pyversion,
-                                                  check_untyped_defs=check_untyped_defs)
+        self.options = options
+        self.semantic_analyzer = SemanticAnalyzer(lib_path, self.errors, options=options)
         self.modules = self.semantic_analyzer.modules
         self.semantic_analyzer_pass3 = ThirdPass(self.modules, self.errors)
-        self.type_checker = TypeChecker(self.errors,
-                                        self.modules,
-                                        self.pyversion,
-                                        DISALLOW_UNTYPED_CALLS in self.flags,
-                                        DISALLOW_UNTYPED_DEFS in self.flags,
-                                        check_untyped_defs,
-                                        WARN_INCOMPLETE_STUB in self.flags)
+        self.type_checker = TypeChecker(self.errors, self.modules, options=options)
         self.missing_modules = set()  # type: Set[str]
+        self.stale_modules = set()  # type: Set[str]
 
     def all_imported_modules_in_file(self,
-                                     file: MypyFile) -> List[Tuple[str, int]]:
+                                     file: MypyFile) -> List[Tuple[int, str, int]]:
         """Find all reachable import statements in a file.
 
-        Return list of tuples (module id, import line number) for all modules
-        imported in file.
+        Return list of tuples (priority, module id, import line number)
+        for all modules imported in file; lower numbers == higher priority.
         """
+
         def correct_rel_imp(imp: Union[ImportFrom, ImportAll]) -> str:
             """Function to correct for relative imports."""
             file_id = file.fullname()
@@ -408,21 +363,28 @@ class BuildManager:
 
             return new_id
 
-        res = []  # type: List[Tuple[str, int]]
+        res = []  # type: List[Tuple[int, str, int]]
         for imp in file.imports:
             if not imp.is_unreachable:
                 if isinstance(imp, Import):
+                    pri = PRI_MED if imp.is_top_level else PRI_LOW
                     for id, _ in imp.ids:
-                        res.append((id, imp.line))
+                        ancestor_parts = id.split(".")[:-1]
+                        ancestors = []
+                        for part in ancestor_parts:
+                            ancestors.append(part)
+                            res.append((PRI_LOW, ".".join(ancestors), imp.line))
+                        res.append((pri, id, imp.line))
                 elif isinstance(imp, ImportFrom):
                     cur_id = correct_rel_imp(imp)
                     pos = len(res)
                     all_are_submodules = True
                     # Also add any imported names that are submodules.
+                    pri = PRI_MED if imp.is_top_level else PRI_LOW
                     for name, __ in imp.names:
                         sub_id = cur_id + '.' + name
                         if self.is_module(sub_id):
-                            res.append((sub_id, imp.line))
+                            res.append((pri, sub_id, imp.line))
                         else:
                             all_are_submodules = False
                     # If all imported names are submodules, don't add
@@ -431,9 +393,12 @@ class BuildManager:
                     # cur_id is also a dependency, and we should
                     # insert it *before* any submodules.
                     if not all_are_submodules:
-                        res.insert(pos, ((cur_id, imp.line)))
+                        pri = PRI_HIGH if imp.is_top_level else PRI_LOW
+                        res.insert(pos, ((pri, cur_id, imp.line)))
                 elif isinstance(imp, ImportAll):
-                    res.append((correct_rel_imp(imp), imp.line))
+                    pri = PRI_HIGH if imp.is_top_level else PRI_LOW
+                    res.append((pri, correct_rel_imp(imp), imp.line))
+
         return res
 
     def is_module(self, id: str) -> bool:
@@ -446,21 +411,35 @@ class BuildManager:
         Raise CompileError if there is a parse error.
         """
         num_errs = self.errors.num_messages()
-        tree = parse(source, path, self.errors,
-                     pyversion=self.pyversion,
-                     custom_typing_module=self.custom_typing_module,
-                     fast_parser=FAST_PARSER in self.flags)
+        tree = parse(source, path, self.errors, options=self.options)
         tree._fullname = id
+
+        # We don't want to warn about 'type: ignore' comments on
+        # imports, but we're about to modify tree.imports, so grab
+        # these first.
+        import_lines = set(node.line for node in tree.imports)
+
+        # Skip imports that have been ignored (so that we can ignore a C extension module without
+        # stub, for example), except for 'from x import *', because we wouldn't be able to
+        # determine which names should be defined unless we process the module. We can still
+        # ignore errors such as redefinitions when using the latter form.
+        imports = [node for node in tree.imports
+                   if node.line not in tree.ignored_lines or isinstance(node, ImportAll)]
+        tree.imports = imports
+
         if self.errors.num_messages() != num_errs:
             self.log("Bailing due to parse errors")
             self.errors.raise_error()
+
+        self.errors.set_file_ignored_lines(path, tree.ignored_lines)
+        self.errors.mark_file_ignored_lines_used(path, import_lines)
         return tree
 
     def module_not_found(self, path: str, line: int, id: str) -> None:
         self.errors.set_file(path)
         stub_msg = "(Stub files are from https://github.com/python/typeshed)"
-        if ((self.pyversion[0] == 2 and moduleinfo.is_py2_std_lib_module(id)) or
-                (self.pyversion[0] >= 3 and moduleinfo.is_py3_std_lib_module(id))):
+        if ((self.options.python_version[0] == 2 and moduleinfo.is_py2_std_lib_module(id)) or
+                (self.options.python_version[0] >= 3 and moduleinfo.is_py3_std_lib_module(id))):
             self.errors.report(
                 line, "No library stub file for standard library module '{}'".format(id))
             self.errors.report(line, stub_msg, severity='note', only_once=True)
@@ -478,12 +457,12 @@ class BuildManager:
             self.reports.file(file, type_map=self.type_checker.type_map)
 
     def log(self, *message: str) -> None:
-        if VERBOSE in self.flags:
+        if self.options.verbosity >= 1:
             print('%.3f:LOG: ' % (time.time() - self.start_time), *message, file=sys.stderr)
             sys.stderr.flush()
 
     def trace(self, *message: str) -> None:
-        if self.flags.count(VERBOSE) >= 2:
+        if self.options.verbosity >= 2:
             print('%.3f:TRACE:' % (time.time() - self.start_time), *message, file=sys.stderr)
             sys.stderr.flush()
 
@@ -533,7 +512,7 @@ find_module_dir_cache = {}  # type: Dict[Tuple[str, Tuple[str, ...]], List[str]]
 find_module_listdir_cache = {}  # type: Dict[str, Optional[List[str]]]
 
 
-def find_module_clear_caches():
+def find_module_clear_caches() -> None:
     find_module_cache.clear()
     find_module_dir_cache.clear()
     find_module_listdir_cache.clear()
@@ -572,12 +551,11 @@ def is_file(path: str) -> bool:
     return os.path.isfile(path)
 
 
-def find_module(id: str, lib_path: Iterable[str]) -> str:
+def find_module(id: str, lib_path_arg: Iterable[str]) -> str:
     """Return the path of the module source file, or None if not found."""
-    if not isinstance(lib_path, tuple):
-        lib_path = tuple(lib_path)
+    lib_path = tuple(lib_path_arg)
 
-    def find():
+    def find() -> Optional[str]:
         # If we're looking for a module like 'foo.bar.baz', it's likely that most of the
         # many elements of lib_path don't even have a subdirectory 'foo/bar'.  Discover
         # that only once and cache it for when we look for modules like 'foo.bar.blah'
@@ -693,22 +671,21 @@ def read_with_python_encoding(path: str, pyversion: Tuple[int, int]) -> str:
         return source_bytearray.decode(encoding)
 
 
-MYPY_CACHE = '.mypy_cache'
-
-
-def get_cache_names(id: str, path: str, pyversion: Tuple[int, int]) -> Tuple[str, str]:
+def get_cache_names(id: str, path: str, cache_dir: str,
+                    pyversion: Tuple[int, int]) -> Tuple[str, str]:
     """Return the file names for the cache files.
 
     Args:
       id: module ID
       path: module path (used to recognize packages)
+      cache_dir: cache directory
       pyversion: Python version (major, minor)
 
     Returns:
       A tuple with the file names to be used for the meta JSON and the
       data JSON, respectively.
     """
-    prefix = os.path.join(MYPY_CACHE, '%d.%d' % pyversion, *id.split('.'))
+    prefix = os.path.join(cache_dir, '%d.%d' % pyversion, *id.split('.'))
     is_package = os.path.basename(path).startswith('__init__.py')
     if is_package:
         prefix = os.path.join(prefix, '__init__')
@@ -728,7 +705,8 @@ def find_cache_meta(id: str, path: str, manager: BuildManager) -> Optional[Cache
       valid; otherwise None.
     """
     # TODO: May need to take more build options into account
-    meta_json, data_json = get_cache_names(id, path, manager.pyversion)
+    meta_json, data_json = get_cache_names(
+        id, path, manager.options.cache_dir, manager.options.python_version)
     manager.trace('Looking for {} {}'.format(id, data_json))
     if not os.path.exists(meta_json):
         return None
@@ -748,20 +726,22 @@ def find_cache_meta(id: str, path: str, manager: BuildManager) -> Optional[Cache
         meta.get('data_mtime'),
         data_json,
         meta.get('suppressed', []),
-        meta.get('flags'),
+        meta.get('options'),
+        meta.get('dep_prios', []),
     )
     if (m.id != id or m.path != path or
             m.mtime is None or m.size is None or
             m.dependencies is None or m.data_mtime is None):
         return None
 
-    # Metadata generated by older mypy version and no flags were saved
-    if m.flags is None:
+    # Ignore cache if generated by an older mypy version.
+    if m.options is None or len(m.dependencies) != len(m.dep_prios):
         return None
 
-    cached_flags = select_flags_affecting_cache(m.flags)
-    current_flags = select_flags_affecting_cache(manager.flags)
-    if cached_flags != current_flags:
+    # Ignore cache if (relevant) options aren't the same.
+    cached_options = m.options
+    current_options = select_options_affecting_cache(manager.options)
+    if cached_options != current_options:
         return None
 
     # TODO: Share stat() outcome with find_module()
@@ -778,25 +758,26 @@ def find_cache_meta(id: str, path: str, manager: BuildManager) -> Optional[Cache
     return m
 
 
-def select_flags_affecting_cache(flags: Iterable[str]) -> AbstractSet[str]:
-    return set(flags).intersection(FLAGS_AFFECTING_CACHE)
+def select_options_affecting_cache(options: Options) -> Mapping[str, bool]:
+    return {opt: getattr(options, opt) for opt in OPTIONS_AFFECTING_CACHE}
 
 
-FLAGS_AFFECTING_CACHE = set([
-    SILENT_IMPORTS,
-    ALMOST_SILENT,
-    DISALLOW_UNTYPED_CALLS,
-    DISALLOW_UNTYPED_DEFS,
-    CHECK_UNTYPED_DEFS,
-])
+OPTIONS_AFFECTING_CACHE = [
+    "silent_imports",
+    "almost_silent",
+    "disallow_untyped_calls",
+    "disallow_untyped_defs",
+    "check_untyped_defs",
+]
 
 
-def random_string():
+def random_string() -> str:
     return binascii.hexlify(os.urandom(8)).decode('ascii')
 
 
 def write_cache(id: str, path: str, tree: MypyFile,
                 dependencies: List[str], suppressed: List[str],
+                dep_prios: List[int],
                 manager: BuildManager) -> None:
     """Write cache files for a module.
 
@@ -806,6 +787,7 @@ def write_cache(id: str, path: str, tree: MypyFile,
       tree: the fully checked module data
       dependencies: module IDs on which this module depends
       suppressed: module IDs which were suppressed as dependencies
+      dep_prios: priorities (parallel array to dependencies)
       manager: the build manager (for pyversion, log/trace)
     """
     path = os.path.abspath(path)
@@ -813,7 +795,8 @@ def write_cache(id: str, path: str, tree: MypyFile,
     st = os.stat(path)  # TODO: Errors
     mtime = st.st_mtime
     size = st.st_size
-    meta_json, data_json = get_cache_names(id, path, manager.pyversion)
+    meta_json, data_json = get_cache_names(
+        id, path, manager.options.cache_dir, manager.options.python_version)
     manager.log('Writing {} {} {}'.format(id, meta_json, data_json))
     data = tree.serialize()
     parent = os.path.dirname(data_json)
@@ -834,19 +817,14 @@ def write_cache(id: str, path: str, tree: MypyFile,
             'data_mtime': data_mtime,
             'dependencies': dependencies,
             'suppressed': suppressed,
-            'flags': manager.flags,
+            'options': select_options_affecting_cache(manager.options),
+            'dep_prios': dep_prios,
             }
     with open(meta_json_tmp, 'w') as f:
         json.dump(meta, f, sort_keys=True)
         f.write('\n')
-    # TODO: This is a temporary change until Python 3.2 support is dropped, see #1504
-    # os.rename will raise an exception rather than replace files on Windows
-    try:
-        replace = os.replace
-    except AttributeError:
-        replace = os.rename
-    replace(data_json_tmp, data_json)
-    replace(meta_json_tmp, meta_json)
+    os.replace(data_json_tmp, data_json)
+    os.replace(meta_json_tmp, meta_json)
 
 
 """Dependency manager.
@@ -1013,6 +991,7 @@ class State:
     tree = None  # type: Optional[MypyFile]
     dependencies = None  # type: List[str]
     suppressed = None  # type: List[str]  # Suppressed/missing dependencies
+    priorities = None  # type: Dict[str, int]
 
     # Map each dependency to the line number where it is first imported
     dep_line_map = None  # type: Dict[str, int]
@@ -1052,7 +1031,7 @@ class State:
         self.id = id or '__main__'
         if not path and source is None:
             file_id = id
-            if id == 'builtins' and manager.pyversion[0] == 2:
+            if id == 'builtins' and manager.options.python_version[0] == 2:
                 # The __builtin__ module is called internally by mypy
                 # 'builtins' in Python 2 mode (similar to Python 3),
                 # but the stub file is __builtin__.pyi.  The reason is
@@ -1065,7 +1044,7 @@ class State:
             path = find_module(file_id, manager.lib_path)
             if path:
                 # In silent mode, don't import .py files, except from stubs.
-                if (SILENT_IMPORTS in manager.flags and
+                if (manager.options.silent_imports and
                         path.endswith('.py') and (caller_state or ancestor_for)):
                     # (Never silence builtins, even if it's a .py file;
                     # this can happen in tests!)
@@ -1073,7 +1052,7 @@ class State:
                         not ((caller_state and
                               caller_state.tree and
                               caller_state.tree.is_stub))):
-                        if ALMOST_SILENT in manager.flags:
+                        if manager.options.almost_silent:
                             if ancestor_for:
                                 self.skipping_ancestor(id, path, ancestor_for)
                             else:
@@ -1086,11 +1065,10 @@ class State:
                 # misspelled module name, missing stub, module not in
                 # search path or the module has not been installed.
                 if caller_state:
-                    suppress_message = ((SILENT_IMPORTS in manager.flags and
-                                        ALMOST_SILENT not in manager.flags) or
+                    suppress_message = ((manager.options.silent_imports and
+                                        not manager.options.almost_silent) or
                                         (caller_state.tree is not None and
-                                         (caller_line in caller_state.tree.ignored_lines or
-                                          'import' in caller_state.tree.weak_opts)))
+                                         'import' in caller_state.tree.weak_opts))
                     if not suppress_message:
                         save_import_context = manager.errors.import_context()
                         manager.errors.set_import_context(caller_state.import_context)
@@ -1106,7 +1084,7 @@ class State:
         self.path = path
         self.xpath = path or '<string>'
         self.source = source
-        if path and source is None and INCREMENTAL in manager.flags:
+        if path and source is None and manager.options.incremental:
             self.meta = find_cache_meta(self.id, self.path, manager)
             # TODO: Get mtime if not cached.
         self.add_ancestors()
@@ -1115,6 +1093,9 @@ class State:
             # compare them to the originals later.
             self.dependencies = list(self.meta.dependencies)
             self.suppressed = list(self.meta.suppressed)
+            assert len(self.meta.dependencies) == len(self.meta.dep_prios)
+            self.priorities = {id: pri
+                               for id, pri in zip(self.meta.dependencies, self.meta.dep_prios)}
             self.dep_line_map = {}
         else:
             # Parse the file (and then some) to get the dependencies.
@@ -1171,6 +1152,7 @@ class State:
     def mark_stale(self) -> None:
         """Throw away the cache data for this file, marking it as stale."""
         self.meta = None
+        self.manager.stale_modules.add(self.id)
 
     def check_blockers(self) -> None:
         """Raise CompileError if a blocking error is detected."""
@@ -1222,7 +1204,7 @@ class State:
             self.source = None  # We won't need it again.
             if self.path and source is None:
                 try:
-                    source = read_with_python_encoding(self.path, manager.pyversion)
+                    source = read_with_python_encoding(self.path, manager.options.python_version)
                 except IOError as ioerr:
                     raise CompileError([
                         "mypy: can't read file '{}': {}".format(self.path, ioerr.strerror)])
@@ -1250,8 +1232,10 @@ class State:
         # Also keep track of each dependency's source line.
         dependencies = []
         suppressed = []
+        priorities = {}  # type: Dict[str, int]  # id -> priority
         dep_line_map = {}  # type: Dict[str, int]  # id -> line
-        for id, line in manager.all_imported_modules_in_file(self.tree):
+        for pri, id, line in manager.all_imported_modules_in_file(self.tree):
+            priorities[id] = min(pri, priorities.get(id, PRI_ALL))
             if id == self.id:
                 continue
             # Omit missing modules, as otherwise we could not type-check
@@ -1282,6 +1266,7 @@ class State:
         # for differences (e.g. --silent-imports).
         self.dependencies = dependencies
         self.suppressed = suppressed
+        self.priorities = priorities
         self.dep_line_map = dep_line_map
         self.check_blockers()
 
@@ -1305,24 +1290,26 @@ class State:
     def semantic_analysis_pass_three(self) -> None:
         with self.wrap_context():
             self.manager.semantic_analyzer_pass3.visit_file(self.tree, self.xpath)
-            if DUMP_TYPE_STATS in self.manager.flags:
+            if self.manager.options.dump_type_stats:
                 dump_type_stats(self.tree, self.xpath)
 
     def type_check(self) -> None:
         manager = self.manager
-        if manager.target < TYPE_CHECK:
+        if manager.options.semantic_analysis_only:
             return
         with self.wrap_context():
             manager.type_checker.visit_file(self.tree, self.xpath)
-            if DUMP_INFER_STATS in manager.flags:
+            if manager.options.dump_inference_stats:
                 dump_type_stats(self.tree, self.xpath, inferred=True,
                                 typemap=manager.type_checker.type_map)
             manager.report_file(self.tree)
 
     def write_cache(self) -> None:
-        if self.path and INCREMENTAL in self.manager.flags and not self.manager.errors.is_errors():
+        if self.path and self.manager.options.incremental and not self.manager.errors.is_errors():
+            dep_prios = [self.priorities.get(dep, PRI_HIGH) for dep in self.dependencies]
             write_cache(self.id, self.path, self.tree,
                         list(self.dependencies), list(self.suppressed),
+                        dep_prios,
                         self.manager)
 
 
@@ -1334,6 +1321,8 @@ def dispatch(sources: List[BuildSource], manager: BuildManager) -> None:
     graph = load_graph(sources, manager)
     manager.log("Loaded graph with %d nodes" % len(graph))
     process_graph(graph, manager)
+    if manager.options.warn_unused_ignores:
+        manager.errors.generate_unused_ignore_notes()
 
 
 def load_graph(sources: List[BuildSource], manager: BuildManager) -> Graph:
@@ -1351,7 +1340,7 @@ def load_graph(sources: List[BuildSource], manager: BuildManager) -> Graph:
             continue
         if st.id in graph:
             manager.errors.set_file(st.xpath)
-            manager.errors.report(1, "Duplicate module named '%s'" % st.id)
+            manager.errors.report(-1, "Duplicate module named '%s'" % st.id)
             manager.errors.raise_error()
         graph[st.id] = st
         new.append(st)
@@ -1389,10 +1378,9 @@ def process_graph(graph: Graph, manager: BuildManager) -> None:
     # dependencies) to roots (those from which everything else can be
     # reached).
     for ascc in sccs:
-        # Sort the SCC's nodes in *reverse* order or encounter.
-        # This is a heuristic for handling import cycles.
+        # Order the SCC's nodes using a heuristic.
         # Note that ascc is a set, and scc is a list.
-        scc = sorted(ascc, key=lambda id: -graph[id].order)
+        scc = order_ascc(graph, ascc)
         # If builtins is in the list, move it last.  (This is a bit of
         # a hack, but it's necessary because the builtins module is
         # part of a small cycle involving at least {builtins, abc,
@@ -1401,6 +1389,12 @@ def process_graph(graph: Graph, manager: BuildManager) -> None:
         if 'builtins' in ascc:
             scc.remove('builtins')
             scc.append('builtins')
+        if manager.options.verbosity >= 2:
+            for id in scc:
+                manager.trace("Priorities for %s:" % id,
+                              " ".join("%s:%d" % (x, graph[id].priorities[x])
+                                       for x in graph[id].dependencies
+                                       if x in ascc and x in graph[id].priorities))
         # Because the SCCs are presented in topological sort order, we
         # don't need to look at dependencies recursively for staleness
         # -- the immediate dependencies are sufficient.
@@ -1427,7 +1421,7 @@ def process_graph(graph: Graph, manager: BuildManager) -> None:
             # cache file is newer than any scc node's cache file.
             oldest_in_scc = min(graph[id].meta.data_mtime for id in scc)
             newest_in_deps = 0 if not deps else max(graph[dep].meta.data_mtime for dep in deps)
-            if manager.flags.count(VERBOSE) >= 2:  # Dump all mtimes for extreme debugging.
+            if manager.options.verbosity >= 3:  # Dump all mtimes for extreme debugging.
                 all_ids = sorted(ascc | deps, key=lambda id: graph[id].meta.data_mtime)
                 for id in all_ids:
                     if id in scc:
@@ -1467,6 +1461,53 @@ def process_graph(graph: Graph, manager: BuildManager) -> None:
             process_stale_scc(graph, scc)
 
 
+def order_ascc(graph: Graph, ascc: AbstractSet[str], pri_max: int = PRI_ALL) -> List[str]:
+    """Come up with the ideal processing order within an SCC.
+
+    Using the priorities assigned by all_imported_modules_in_file(),
+    try to reduce the cycle to a DAG, by omitting arcs representing
+    dependencies of lower priority.
+
+    In the simplest case, if we have A <--> B where A has a top-level
+    "import B" (medium priority) but B only has the reverse "import A"
+    inside a function (low priority), we turn the cycle into a DAG by
+    dropping the B --> A arc, which leaves only A --> B.
+
+    If all arcs have the same priority, we fall back to sorting by
+    reverse global order (the order in which modules were first
+    encountered).
+
+    The algorithm is recursive, as follows: when as arcs of different
+    priorities are present, drop all arcs of the lowest priority,
+    identify SCCs in the resulting graph, and apply the algorithm to
+    each SCC thus found.  The recursion is bounded because at each
+    recursion the spread in priorities is (at least) one less.
+
+    In practice there are only a few priority levels (currently
+    N=3) and in the worst case we just carry out the same algorithm
+    for finding SCCs N times.  Thus the complexity is no worse than
+    the complexity of the original SCC-finding algorithm -- see
+    strongly_connected_components() below for a reference.
+    """
+    if len(ascc) == 1:
+        return [s for s in ascc]
+    pri_spread = set()
+    for id in ascc:
+        state = graph[id]
+        for dep in state.dependencies:
+            if dep in ascc:
+                pri = state.priorities.get(dep, PRI_HIGH)
+                if pri < pri_max:
+                    pri_spread.add(pri)
+    if len(pri_spread) == 1:
+        # Filtered dependencies are uniform -- order by global order.
+        return sorted(ascc, key=lambda id: -graph[id].order)
+    pri_max = max(pri_spread)
+    sccs = sorted_components(graph, ascc, pri_max)
+    # The recursion is bounded by the len(pri_spread) check above.
+    return [s for ss in sccs for s in order_ascc(graph, ss, pri_max)]
+
+
 def process_fresh_scc(graph: Graph, scc: List[str]) -> None:
     """Process the modules in one SCC from their cached data."""
     for id in scc:
@@ -1498,7 +1539,9 @@ def process_stale_scc(graph: Graph, scc: List[str]) -> None:
         graph[id].write_cache()
 
 
-def sorted_components(graph: Graph) -> List[AbstractSet[str]]:
+def sorted_components(graph: Graph,
+                      vertices: Optional[AbstractSet[str]] = None,
+                      pri_max: int = PRI_ALL) -> List[AbstractSet[str]]:
     """Return the graph's SCCs, topologically sorted by dependencies.
 
     The sort order is from leaves (nodes without dependencies) to
@@ -1508,9 +1551,9 @@ def sorted_components(graph: Graph) -> List[AbstractSet[str]]:
     dependencies that aren't present in graph.keys() are ignored.
     """
     # Compute SCCs.
-    vertices = set(graph)
-    edges = {id: [dep for dep in st.dependencies if dep in graph]
-             for id, st in graph.items()}
+    if vertices is None:
+        vertices = set(graph)
+    edges = {id: deps_filtered(graph, vertices, id, pri_max) for id in vertices}
     sccs = list(strongly_connected_components(vertices, edges))
     # Topsort.
     sccsmap = {id: frozenset(scc) for scc in sccs for id in scc}
@@ -1518,7 +1561,7 @@ def sorted_components(graph: Graph) -> List[AbstractSet[str]]:
     for scc in sccs:
         deps = set()  # type: Set[AbstractSet[str]]
         for id in scc:
-            deps.update(sccsmap[x] for x in graph[id].dependencies if x in graph)
+            deps.update(sccsmap[x] for x in deps_filtered(graph, vertices, id, pri_max))
         data[frozenset(scc)] = deps
     res = []
     for ready in topsort(data):
@@ -1535,7 +1578,17 @@ def sorted_components(graph: Graph) -> List[AbstractSet[str]]:
     return res
 
 
-def strongly_connected_components(vertices: Set[str],
+def deps_filtered(graph: Graph, vertices: AbstractSet[str], id: str, pri_max: int) -> List[str]:
+    """Filter dependencies for id with pri < pri_max."""
+    if id not in vertices:
+        return []
+    state = graph[id]
+    return [dep
+            for dep in state.dependencies
+            if dep in vertices and state.priorities.get(dep, PRI_HIGH) < pri_max]
+
+
+def strongly_connected_components(vertices: AbstractSet[str],
                                   edges: Dict[str, List[str]]) -> Iterator[Set[str]]:
     """Compute Strongly Connected Components of a directed graph.
 
