@@ -4,8 +4,8 @@ from typing import cast, Dict, List, Tuple, Callable, Union, Optional
 
 from mypy.types import (
     Type, AnyType, CallableType, Overloaded, NoneTyp, Void, TypeVarDef,
-    TupleType, Instance, TypeVarType, ErasedType, UnionType,
-    PartialType, DeletedType, UnboundType, TypeType
+    TupleType, Instance, TypeVarId, TypeVarType, ErasedType, UnionType,
+    PartialType, DeletedType, UnboundType, UninhabitedType, TypeType
 )
 from mypy.nodes import (
     NameExpr, RefExpr, Var, FuncDef, OverloadedFuncDef, TypeInfo, CallExpr,
@@ -22,7 +22,7 @@ from mypy import nodes
 import mypy.checker
 from mypy import types
 from mypy.sametypes import is_same_type
-from mypy.replacetvars import replace_func_type_vars
+from mypy.erasetype import replace_meta_vars
 from mypy.messages import MessageBuilder
 from mypy import messages
 from mypy.infer import infer_type_arguments, infer_function_type_arguments
@@ -34,7 +34,10 @@ from mypy.checkmember import analyze_member_access, type_object_type
 from mypy.semanal import self_type
 from mypy.constraints import get_actual_type
 from mypy.checkstrformat import StringFormatterChecker
+from mypy.expandtype import expand_type
+import mypy.checkexpr
 
+from mypy import experiments
 
 # Type of callback user for checking individual function arguments. See
 # check_args() below for details.
@@ -133,7 +136,7 @@ class ExpressionChecker:
             return self.accept(e.analyzed, self.chk.type_context[-1])
         self.try_infer_partial_type(e)
         callee_type = self.accept(e.callee)
-        if (self.chk.disallow_untyped_calls and
+        if (self.chk.options.disallow_untyped_calls and
                 self.chk.typing_mode_full() and
                 isinstance(callee_type, CallableType)
                 and callee_type.implicit):
@@ -154,19 +157,21 @@ class ExpressionChecker:
             var = cast(Var, e.callee.expr.node)
             partial_types = self.chk.find_partial_types(var)
             if partial_types is not None and not self.chk.current_node_deferred:
-                partial_type_type = cast(PartialType, var.type).type
-                if partial_type_type is None:
+                partial_type = cast(PartialType, var.type)
+                if partial_type is None or partial_type.type is None:
                     # A partial None type -> can't infer anything.
                     return
-                typename = partial_type_type.fullname()
+                typename = partial_type.type.fullname()
                 methodname = e.callee.name
                 # Sometimes we can infer a full type for a partial List, Dict or Set type.
                 # TODO: Don't infer argument expression twice.
                 if (typename in self.item_args and methodname in self.item_args[typename]
                         and e.arg_kinds == [ARG_POS]):
                     item_type = self.accept(e.args[0])
-                    if mypy.checker.is_valid_inferred_type(item_type):
-                        var.type = self.chk.named_generic_type(typename, [item_type])
+                    full_item_type = UnionType.make_simplified_union(
+                        [item_type, partial_type.inner_types[0]])
+                    if mypy.checker.is_valid_inferred_type(full_item_type):
+                        var.type = self.chk.named_generic_type(typename, [full_item_type])
                         del partial_types[var]
                 elif (typename in self.container_args
                       and methodname in self.container_args[typename]
@@ -175,10 +180,15 @@ class ExpressionChecker:
                     if isinstance(arg_type, Instance):
                         arg_typename = arg_type.type.fullname()
                         if arg_typename in self.container_args[typename][methodname]:
+                            full_item_types = [
+                                UnionType.make_simplified_union([item_type, prev_type])
+                                for item_type, prev_type
+                                in zip(arg_type.args, partial_type.inner_types)
+                            ]
                             if all(mypy.checker.is_valid_inferred_type(item_type)
-                                   for item_type in arg_type.args):
+                                   for item_type in full_item_types):
                                 var.type = self.chk.named_generic_type(typename,
-                                                                       list(arg_type.args))
+                                                                       list(full_item_types))
                                 del partial_types[var]
 
     def check_call_expr_with_callee_type(self, callee_type: Type,
@@ -226,6 +236,7 @@ class ExpressionChecker:
                 lambda i: self.accept(args[i]))
 
             if callee.is_generic():
+                callee = freshen_generic_callable(callee)
                 callee = self.infer_function_type_arguments_using_context(
                     callee, context)
                 callee = self.infer_function_type_arguments(
@@ -240,6 +251,11 @@ class ExpressionChecker:
             self.check_argument_types(arg_types, arg_kinds, callee,
                                       formal_to_actual, context,
                                       messages=arg_messages)
+
+            if (callee.is_type_obj() and (len(arg_types) == 1)
+                    and is_equivalent(callee.ret_type, self.named_type('builtins.type'))):
+                callee = callee.copy_modified(ret_type=TypeType(arg_types[0]))
+
             if callable_node:
                 # Store the inferred callable type.
                 self.chk.store_type(callable_node, callee)
@@ -270,8 +286,8 @@ class ExpressionChecker:
                     callee)
         elif isinstance(callee, Instance):
             call_function = analyze_member_access('__call__', callee, context,
-                                         False, False, self.named_type, self.not_ready_callback,
-                                         self.msg)
+                                         False, False, False, self.named_type,
+                                         self.not_ready_callback, self.msg)
             return self.check_call(call_function, args, arg_kinds, context, arg_names,
                                    callable_node, arg_messages)
         elif isinstance(callee, TypeVarType):
@@ -320,7 +336,7 @@ class ExpressionChecker:
         self.msg.unsupported_type_type(item, context)
         return AnyType()
 
-    def infer_arg_types_in_context(self, callee: CallableType,
+    def infer_arg_types_in_context(self, callee: Optional[CallableType],
                                    args: List[Node]) -> List[Type]:
         """Infer argument expression types using a callable type as context.
 
@@ -386,16 +402,16 @@ class ExpressionChecker:
         ctx = self.chk.type_context[-1]
         if not ctx:
             return callable
-        # The return type may have references to function type variables that
+        # The return type may have references to type metavariables that
         # we are inferring right now. We must consider them as indeterminate
         # and they are not potential results; thus we replace them with the
         # special ErasedType type. On the other hand, class type variables are
         # valid results.
-        erased_ctx = replace_func_type_vars(ctx, ErasedType())
+        erased_ctx = replace_meta_vars(ctx, ErasedType())
         ret_type = callable.ret_type
         if isinstance(ret_type, TypeVarType):
             if ret_type.values or (not isinstance(ctx, Instance) or
-                                   not cast(Instance, ctx).args):
+                                   not ctx.args):
                 # The return type is a type variable. If it has values, we can't easily restrict
                 # type inference to conform to the valid values. If it's unrestricted, we could
                 # infer a too general type for the type variable if we use context, and this could
@@ -411,7 +427,7 @@ class ExpressionChecker:
         # Only substitute non-None and non-erased types.
         new_args = []  # type: List[Type]
         for arg in args:
-            if isinstance(arg, NoneTyp) or has_erased_component(arg):
+            if isinstance(arg, (NoneTyp, UninhabitedType)) or has_erased_component(arg):
                 new_args.append(None)
             else:
                 new_args.append(arg)
@@ -470,7 +486,7 @@ class ExpressionChecker:
                 #       if they shuffle type variables around, as we assume that there is a 1-1
                 #       correspondence with dict type variables. This is a marginal issue and
                 #       a little tricky to fix so it's left unfixed for now.
-                if isinstance(inferred_args[0], NoneTyp):
+                if isinstance(inferred_args[0], (NoneTyp, UninhabitedType)):
                     inferred_args[0] = self.named_type('builtins.str')
                 elif not is_subtype(self.named_type('builtins.str'), inferred_args[0]):
                     self.msg.fail(messages.KEYWORD_ARGUMENT_REQUIRES_STR_KEY_TYPE,
@@ -504,7 +520,7 @@ class ExpressionChecker:
         # information to infer the argument. Replace them with None values so
         # that they are not applied yet below.
         for i, arg in enumerate(inferred_args):
-            if isinstance(arg, NoneTyp) or has_erased_component(arg):
+            if isinstance(arg, (NoneTyp, UninhabitedType)) or has_erased_component(arg):
                 inferred_args[i] = None
 
         callee_type = cast(CallableType, self.apply_generic_arguments(
@@ -845,7 +861,7 @@ class ExpressionChecker:
         else:
             # This is a reference to a non-module attribute.
             return analyze_member_access(e.name, self.accept(e.expr), e,
-                                         is_lvalue, False,
+                                         is_lvalue, False, False,
                                          self.named_type, self.not_ready_callback, self.msg)
 
     def analyze_external_member_access(self, member: str, base_type: Type,
@@ -854,7 +870,7 @@ class ExpressionChecker:
         refer to private definitions. Return the result type.
         """
         # TODO remove; no private definitions in mypy
-        return analyze_member_access(member, base_type, context, False, False,
+        return analyze_member_access(member, base_type, context, False, False, False,
                                      self.named_type, self.not_ready_callback, self.msg)
 
     def visit_int_expr(self, e: IntExpr) -> Type:
@@ -883,7 +899,7 @@ class ExpressionChecker:
 
     def visit_ellipsis(self, e: EllipsisExpr) -> Type:
         """Type check '...'."""
-        if self.chk.pyversion[0] >= 3:
+        if self.chk.options.python_version[0] >= 3:
             return self.named_type('builtins.ellipsis')
         else:
             # '...' is not valid in normal Python 2 code, but it can
@@ -980,7 +996,7 @@ class ExpressionChecker:
         return result
 
     def get_operator_method(self, op: str) -> str:
-        if op == '/' and self.chk.pyversion[0] == 2:
+        if op == '/' and self.chk.options.python_version[0] == 2:
             # TODO also check for "from __future__ import division"
             return '__div__'
         else:
@@ -992,7 +1008,7 @@ class ExpressionChecker:
 
         Return tuple (result type, inferred operator method type).
         """
-        method_type = analyze_member_access(method, base_type, context, False, False,
+        method_type = analyze_member_access(method, base_type, context, False, False, True,
                                             self.named_type, self.not_ready_callback, local_errors)
         return self.check_call(method_type, [arg], [nodes.ARG_POS],
                                context, arg_messages=local_errors)
@@ -1052,7 +1068,7 @@ class ExpressionChecker:
                                            self.msg)
 
     def get_reverse_op_method(self, method: str) -> str:
-        if method == '__div__' and self.chk.pyversion[0] == 2:
+        if method == '__div__' and self.chk.options.python_version[0] == 2:
             return '__rdiv__'
         else:
             return nodes.reverse_op_methods[method]
@@ -1070,24 +1086,38 @@ class ExpressionChecker:
         left_type = self.accept(e.left, ctx)
 
         if e.op == 'and':
-            # else_map unused
-            if_map, else_map = \
+            right_map, left_map = \
+                mypy.checker.find_isinstance_check(e.left, self.chk.type_map,
+                                                   self.chk.typing_mode_weak())
+        elif e.op == 'or':
+            left_map, right_map = \
                 mypy.checker.find_isinstance_check(e.left, self.chk.type_map,
                                                    self.chk.typing_mode_weak())
         else:
-            if_map = None
+            left_map = None
+            right_map = None
 
-        self.chk.binder.push_frame()
-        if if_map:
-            for var, type in if_map.items():
-                self.chk.binder.push(var, type)
+        if left_map and e.left in left_map:
+            # The type of expressions in left_map is the type they'll have if
+            # the left operand is the result of the operator.
+            left_type = left_map[e.left]
 
-        right_type = self.accept(e.right, left_type)
+        with self.chk.binder.frame_context():
+            if right_map:
+                for var, type in right_map.items():
+                    self.chk.binder.push(var, type)
 
-        self.chk.binder.pop_frame()
+            right_type = self.accept(e.right, left_type)
 
         self.check_not_void(left_type, context)
         self.check_not_void(right_type, context)
+
+        # If either of the type maps is None that means that result cannot happen.
+        # If both of the type maps are None we just have no information.
+        if left_map is not None and right_map is None:
+            return left_type
+        elif left_map is None and right_map is not None:
+            return right_type
         return UnionType.make_simplified_union([left_type, right_type])
 
     def check_list_multiply(self, e: OpExpr) -> Type:
@@ -1175,7 +1205,7 @@ class ExpressionChecker:
             e.method_type = method_type
             return result
 
-    def visit_tuple_slice_helper(self, left_type: TupleType, slic: SliceExpr):
+    def visit_tuple_slice_helper(self, left_type: TupleType, slic: SliceExpr) -> Type:
         begin = None  # type: int
         end = None  # type: int
         stride = None  # type:int
@@ -1221,6 +1251,8 @@ class ExpressionChecker:
         """Type check a cast expression."""
         source_type = self.accept(expr.expr, context=AnyType())
         target_type = expr.type
+        if self.chk.options.warn_redundant_casts and is_same_type(source_type, target_type):
+            self.msg.redundant_cast(target_type, expr)
         if not self.is_valid_cast(source_type, target_type):
             self.msg.invalid_cast(target_type, source_type, expr)
         return target_type
@@ -1256,7 +1288,8 @@ class ExpressionChecker:
     def check_list_or_set_expr(self, items: List[Node], fullname: str,
                                tag: str, context: Context) -> Type:
         # Translate into type checking a generic function call.
-        tv = TypeVarType('T', -1, [], self.chk.object_type())
+        tvdef = TypeVarDef('T', -1, [], self.chk.object_type())
+        tv = TypeVarType(tvdef)
         constructor = CallableType(
             [tv],
             [nodes.ARG_STAR],
@@ -1264,7 +1297,7 @@ class ExpressionChecker:
             self.chk.named_generic_type(fullname, [tv]),
             self.named_type('builtins.function'),
             name=tag,
-            variables=[TypeVarDef('T', -1, None, self.chk.object_type())])
+            variables=[tvdef])
         return self.check_call(constructor,
                                items,
                                [nodes.ARG_POS] * len(items), context)[0]
@@ -1293,20 +1326,21 @@ class ExpressionChecker:
 
     def visit_dict_expr(self, e: DictExpr) -> Type:
         # Translate into type checking a generic function call.
-        tv1 = TypeVarType('KT', -1, [], self.chk.object_type())
-        tv2 = TypeVarType('VT', -2, [], self.chk.object_type())
+        ktdef = TypeVarDef('KT', -1, [], self.chk.object_type())
+        vtdef = TypeVarDef('VT', -2, [], self.chk.object_type())
+        kt = TypeVarType(ktdef)
+        vt = TypeVarType(vtdef)
         # The callable type represents a function like this:
         #
         #   def <unnamed>(*v: Tuple[kt, vt]) -> Dict[kt, vt]: ...
         constructor = CallableType(
-            [TupleType([tv1, tv2], self.named_type('builtins.tuple'))],
+            [TupleType([kt, vt], self.named_type('builtins.tuple'))],
             [nodes.ARG_STAR],
             [None],
-            self.chk.named_generic_type('builtins.dict', [tv1, tv2]),
+            self.chk.named_generic_type('builtins.dict', [kt, vt]),
             self.named_type('builtins.function'),
             name='<list>',
-            variables=[TypeVarDef('KT', -1, None, self.chk.object_type()),
-                       TypeVarDef('VT', -2, None, self.chk.object_type())])
+            variables=[ktdef, vtdef])
         # Synthesize function arguments.
         args = []  # type: List[Node]
         for key, value in e.items:
@@ -1335,6 +1369,11 @@ class ExpressionChecker:
             if e.expr() not in self.chk.type_map:
                 self.accept(e.expr())
             ret_type = self.chk.type_map[e.expr()]
+            if isinstance(ret_type, NoneTyp):
+                # For "lambda ...: None", just use type from the context.
+                # Important when the context is Callable[..., None] which
+                # really means Void. See #1425.
+                return inferred_type
             return replace_callable_return_type(inferred_type, ret_type)
 
     def infer_lambda_type_using_context(self, e: FuncExpr) -> CallableType:
@@ -1352,7 +1391,7 @@ class ExpressionChecker:
         # they must be considered as indeterminate. We use ErasedType since it
         # does not affect type inference results (it is for purposes like this
         # only).
-        ctx = replace_func_type_vars(ctx, ErasedType())
+        ctx = replace_meta_vars(ctx, ErasedType())
 
         callable_ctx = cast(CallableType, ctx)
 
@@ -1395,7 +1434,7 @@ class ExpressionChecker:
                     if not self.chk.typing_mode_full():
                         return AnyType()
                     return analyze_member_access(e.name, self_type(e.info), e,
-                                                 is_lvalue, True,
+                                                 is_lvalue, True, False,
                                                  self.named_type, self.not_ready_callback,
                                                  self.msg, base)
         else:
@@ -1430,7 +1469,8 @@ class ExpressionChecker:
 
         # Infer the type of the list comprehension by using a synthetic generic
         # callable type.
-        tv = TypeVarType('T', -1, [], self.chk.object_type())
+        tvdef = TypeVarDef('T', -1, [], self.chk.object_type())
+        tv = TypeVarType(tvdef)
         constructor = CallableType(
             [tv],
             [nodes.ARG_POS],
@@ -1438,27 +1478,28 @@ class ExpressionChecker:
             self.chk.named_generic_type(type_name, [tv]),
             self.chk.named_type('builtins.function'),
             name=id_for_messages,
-            variables=[TypeVarDef('T', -1, None, self.chk.object_type())])
+            variables=[tvdef])
         return self.check_call(constructor,
                                [gen.left_expr], [nodes.ARG_POS], gen)[0]
 
-    def visit_dictionary_comprehension(self, e: DictionaryComprehension):
+    def visit_dictionary_comprehension(self, e: DictionaryComprehension) -> Type:
         """Type check a dictionary comprehension."""
         self.check_for_comp(e)
 
         # Infer the type of the list comprehension by using a synthetic generic
         # callable type.
-        key_tv = TypeVarType('KT', -1, [], self.chk.object_type())
-        value_tv = TypeVarType('VT', -2, [], self.chk.object_type())
+        ktdef = TypeVarDef('KT', -1, [], self.chk.object_type())
+        vtdef = TypeVarDef('VT', -2, [], self.chk.object_type())
+        kt = TypeVarType(ktdef)
+        vt = TypeVarType(vtdef)
         constructor = CallableType(
-            [key_tv, value_tv],
+            [kt, vt],
             [nodes.ARG_POS, nodes.ARG_POS],
             [None, None],
-            self.chk.named_generic_type('builtins.dict', [key_tv, value_tv]),
+            self.chk.named_generic_type('builtins.dict', [kt, vt]),
             self.chk.named_type('builtins.function'),
             name='<dictionary-comprehension>',
-            variables=[TypeVarDef('KT', -1, None, self.chk.object_type()),
-                       TypeVarDef('VT', -2, None, self.chk.object_type())])
+            variables=[ktdef, vtdef])
         return self.check_call(constructor,
                                [e.key, e.value], [nodes.ARG_POS, nodes.ARG_POS], e)[0]
 
@@ -1466,14 +1507,13 @@ class ExpressionChecker:
         """Check the for_comp part of comprehensions. That is the part from 'for':
         ... for x in y if z
         """
-        self.chk.binder.push_frame()
-        for index, sequence, conditions in zip(e.indices, e.sequences,
-                                               e.condlists):
-            sequence_type = self.chk.analyze_iterable_item_type(sequence)
-            self.chk.analyze_index_variables(index, sequence_type, e)
-            for condition in conditions:
-                self.accept(condition)
-        self.chk.binder.pop_frame()
+        with self.chk.binder.frame_context():
+            for index, sequence, conditions in zip(e.indices, e.sequences,
+                                                   e.condlists):
+                sequence_type = self.chk.analyze_iterable_item_type(sequence)
+                self.chk.analyze_index_variables(index, sequence_type, e)
+                for condition in conditions:
+                    self.accept(condition)
 
     def visit_conditional_expr(self, e: ConditionalExpr) -> Type:
         cond_type = self.accept(e.cond)
@@ -1512,7 +1552,7 @@ class ExpressionChecker:
 
     def analyze_cond_branch(self, map: Optional[Dict[Node, Type]],
                             node: Node, context: Optional[Type]) -> Type:
-        with self.chk.binder:
+        with self.chk.binder.frame_context():
             if map:
                 for var, type in map.items():
                     self.chk.binder.push(var, type)
@@ -1730,17 +1770,36 @@ def overload_arg_similarity(actual: Type, formal: Type) -> int:
         actual = actual.erase_to_union_or_bound()
     if isinstance(formal, TypeVarType):
         formal = formal.erase_to_union_or_bound()
-    if (isinstance(actual, NoneTyp) or isinstance(actual, AnyType) or
+    if (isinstance(actual, UninhabitedType) or isinstance(actual, AnyType) or
             isinstance(formal, AnyType) or isinstance(formal, CallableType) or
             (isinstance(actual, Instance) and actual.type.fallback_to_any)):
         # These could match anything at runtime.
         return 2
+    if isinstance(actual, NoneTyp):
+        if not experiments.STRICT_OPTIONAL:
+            # NoneTyp matches anything if we're not doing strict Optional checking
+            return 2
+        else:
+            # NoneType is a subtype of object
+            if isinstance(formal, Instance) and formal.type.fullname() == "builtins.object":
+                return 2
     if isinstance(actual, UnionType):
         return max(overload_arg_similarity(item, formal)
                    for item in actual.items)
     if isinstance(formal, UnionType):
         return max(overload_arg_similarity(actual, item)
                    for item in formal.items)
+    if isinstance(formal, TypeType):
+        if isinstance(actual, TypeType):
+            # Since Type[T] is covariant, check if actual = Type[A] is
+            # a subtype of formal = Type[F].
+            return overload_arg_similarity(actual.item, formal.item)
+        elif isinstance(actual, CallableType) and actual.is_type_obj():
+            # Check if the actual is a constructor of some sort.
+            # Note that this is this unsound, since we don't check the __init__ signature.
+            return overload_arg_similarity(actual.ret_type, formal.item)
+        else:
+            return 0
     if isinstance(formal, Instance):
         if isinstance(actual, CallableType):
             actual = actual.fallback
@@ -1757,6 +1816,11 @@ def overload_arg_similarity(actual: Type, formal: Type) -> int:
                 return 1
             else:
                 return 0
+        elif isinstance(actual, TypeType):
+            if formal.type.fullname() in {"builtins.object", "builtins.type"}:
+                return 2
+            else:
+                return 0
         else:
             return 0
     if isinstance(actual, UnboundType) or isinstance(formal, UnboundType):
@@ -1764,3 +1828,14 @@ def overload_arg_similarity(actual: Type, formal: Type) -> int:
         return 2
     # Fall back to a conservative equality check for the remaining kinds of type.
     return 2 if is_same_type(erasetype.erase_type(actual), erasetype.erase_type(formal)) else 0
+
+
+def freshen_generic_callable(callee: CallableType) -> CallableType:
+    tvdefs = []
+    tvmap = {}  # type: Dict[TypeVarId, Type]
+    for v in callee.variables:
+        tvdef = TypeVarDef.new_unification_variable(v)
+        tvdefs.append(tvdef)
+        tvmap[v.id] = TypeVarType(tvdef)
+
+    return cast(CallableType, expand_type(callee, tvmap)).copy_modified(variables=tvdefs)
