@@ -18,6 +18,7 @@ from mypy.nodes import (
 )
 from mypy.types import Type, CallableType, AnyType, UnboundType, TupleType, TypeList, EllipsisType
 from mypy import defaults
+from mypy import experiments
 from mypy.errors import Errors
 
 try:
@@ -35,8 +36,9 @@ except ImportError:
               ' Python 3.3 and greater.')
     sys.exit(1)
 
-T = TypeVar('T')
-U = TypeVar('U')
+T = TypeVar('T', bound=Union[ast35.expr, ast35.stmt])
+U = TypeVar('U', bound=Node)
+V = TypeVar('V')
 
 TYPE_COMMENT_SYNTAX_ERROR = 'syntax error in type comment'
 TYPE_COMMENT_AST_ERROR = 'invalid type comment'
@@ -91,16 +93,16 @@ def parse_type_comment(type_comment: str, line: int) -> Type:
         return TypeConverter(line=line).visit(typ.body)
 
 
-def with_line(f: Callable[[Any, T], U]) -> Callable[[Any, T], U]:
+def with_line(f: Callable[['ASTConverter', T], U]) -> Callable[['ASTConverter', T], U]:
     @wraps(f)
-    def wrapper(self, ast):
+    def wrapper(self: 'ASTConverter', ast: T) -> U:
         node = f(self, ast)
         node.set_line(ast.lineno)
         return node
     return wrapper
 
 
-def find(f: Callable[[T], bool], seq: Sequence[T]) -> T:
+def find(f: Callable[[V], bool], seq: Sequence[V]) -> V:
     for item in seq:
         if f(item):
             return item
@@ -176,7 +178,7 @@ class ASTConverter(ast35.NodeTransformer):
     def as_block(self, stmts: List[ast35.stmt], lineno: int) -> Block:
         b = None
         if stmts:
-            b = Block(self.visit_list(stmts))
+            b = Block(self.fix_function_overloads(self.visit_list(stmts)))
             b.set_line(lineno)
         return b
 
@@ -254,7 +256,8 @@ class ASTConverter(ast35.NodeTransformer):
             # for ellipsis arg
             if (len(func_type_ast.argtypes) == 1 and
                     isinstance(func_type_ast.argtypes[0], ast35.Ellipsis)):
-                arg_types = [AnyType() for a in args]
+                arg_types = [a.type_annotation if a.type_annotation is not None else AnyType()
+                             for a in args]
             else:
                 arg_types = [a if a is not None else AnyType() for
                             a in TypeConverter(line=n.lineno).visit_list(func_type_ast.argtypes)]
@@ -267,6 +270,9 @@ class ASTConverter(ast35.NodeTransformer):
             arg_types = [a.type_annotation for a in args]
             return_type = TypeConverter(line=n.lineno).visit(n.returns)
 
+        if isinstance(return_type, UnboundType):
+            return_type.is_ret_type = True
+
         func_type = None
         if any(arg_types) or return_type:
             func_type = CallableType([a if a is not None else AnyType() for a in arg_types],
@@ -281,6 +287,7 @@ class ASTConverter(ast35.NodeTransformer):
                        func_type)
         if func_type is not None:
             func_type.definition = func_def
+            func_type.line = n.lineno
 
         if n.decorator_list:
             var = Var(func_def.name())
@@ -294,10 +301,20 @@ class ASTConverter(ast35.NodeTransformer):
         else:
             return func_def
 
+    def set_type_optional(self, type: Type, initializer: Node) -> None:
+        if not experiments.STRICT_OPTIONAL:
+            return
+        # Indicate that type should be wrapped in an Optional if arg is initialized to None.
+        optional = isinstance(initializer, NameExpr) and initializer.name == 'None'
+        if isinstance(type, UnboundType):
+            type.optional = optional
+
     def transform_args(self, args: ast35.arguments, line: int) -> List[Argument]:
-        def make_argument(arg, default, kind):
+        def make_argument(arg: ast35.arg, default: Optional[ast35.expr], kind: int) -> Argument:
             arg_type = TypeConverter(line=line).visit(arg.annotation)
-            return Argument(Var(arg.arg), arg_type, self.visit(default), kind)
+            converted_default = self.visit(default)
+            self.set_type_optional(arg_type, converted_default)
+            return Argument(Var(arg.arg), arg_type, converted_default, kind)
 
         new_args = []
         num_no_defaults = len(args.args) - len(args.defaults)
@@ -353,7 +370,7 @@ class ASTConverter(ast35.NodeTransformer):
             metaclass = self.stringify_name(metaclass_arg.value)
 
         cdef = ClassDef(n.name,
-                        Block(self.fix_function_overloads(self.visit_list(n.body))),
+                        self.as_block(n.body, n.lineno),
                         None,
                         self.visit_list(n.bases),
                         metaclass=metaclass)
@@ -516,7 +533,7 @@ class ASTConverter(ast35.NodeTransformer):
             raise RuntimeError('unknown BoolOp ' + str(type(n)))
 
         # potentially inefficient!
-        def group(vals):
+        def group(vals: List[Node]) -> Node:
             if len(vals) == 2:
                 return OpExpr(op, vals[0], vals[1])
             else:
@@ -634,7 +651,7 @@ class ASTConverter(ast35.NodeTransformer):
     # keyword = (identifier? arg, expr value)
     @with_line
     def visit_Call(self, n: ast35.Call) -> Node:
-        def is_star2arg(k):
+        def is_star2arg(k: ast35.keyword) -> bool:
             return k.arg is None
 
         arg_types = self.visit_list(
