@@ -36,6 +36,9 @@ from mypy.types import Type, CallableType, AnyType, UnboundType
 from mypy.parsetype import (
     parse_type, parse_types, parse_signature, TypeParseError, parse_str_as_signature
 )
+from mypy.options import Options
+
+from mypy import experiments
 
 
 precedence = {
@@ -68,30 +71,30 @@ op_comp = set([
 none = Token('')  # Empty token
 
 
-def parse(source: Union[str, bytes], fnam: str = None, errors: Errors = None,
-          pyversion: Tuple[int, int] = defaults.PYTHON3_VERSION,
-          custom_typing_module: str = None,
-          fast_parser: bool = False) -> MypyFile:
+def parse(source: Union[str, bytes],
+          fnam: str,
+          errors: Errors,
+          options: Options) -> MypyFile:
     """Parse a source file, without doing any semantic analysis.
 
     Return the parse tree. If errors is not provided, raise ParseError
     on failure. Otherwise, use the errors object to report parse errors.
 
-    The pyversion (major, minor) argument determines the Python syntax variant.
+    The python_version (major, minor) option determines the Python syntax variant.
     """
-    if fast_parser:
+    if options.fast_parser:
         import mypy.fastparse
         return mypy.fastparse.parse(source,
                                     fnam=fnam,
                                     errors=errors,
-                                    pyversion=pyversion,
-                                    custom_typing_module=custom_typing_module)
+                                    pyversion=options.python_version,
+                                    custom_typing_module=options.custom_typing_module)
 
     is_stub_file = bool(fnam) and fnam.endswith('.pyi')
     parser = Parser(fnam,
                     errors,
-                    pyversion,
-                    custom_typing_module,
+                    options.python_version,
+                    options.custom_typing_module,
                     is_stub_file=is_stub_file)
     tree = parser.parse(source)
     tree.path = fnam
@@ -174,13 +177,7 @@ class Parser:
         defs = self.parse_defs()
         weak_opts = self.weak_opts()
         self.expect_type(Eof)
-        # Skip imports that have been ignored (so that we can ignore a C extension module without
-        # stub, for example), except for 'from x import *', because we wouldn't be able to
-        # determine which names should be defined unless we process the module. We can still
-        # ignore errors such as redefinitions when using the latter form.
-        imports = [node for node in self.imports
-                   if node.line not in self.ignored_lines or isinstance(node, ImportAll)]
-        node = MypyFile(defs, imports, is_bom, self.ignored_lines,
+        node = MypyFile(defs, self.imports, is_bom, self.ignored_lines,
                         weak_opts=weak_opts)
         return node
 
@@ -482,6 +479,9 @@ class Parser:
             if is_error:
                 return None
 
+            if typ and isinstance(typ.ret_type, UnboundType):
+                typ.ret_type.is_ret_type = True
+
             node = FuncDef(name, args, body, typ)
             node.set_line(def_tok)
             if typ is not None:
@@ -782,8 +782,18 @@ class Parser:
             else:
                 kind = nodes.ARG_POS
 
+        self.set_type_optional(type, initializer)
+
         return Argument(variable, type, initializer, kind), require_named
 
+    def set_type_optional(self, type: Type, initializer: Node) -> None:
+        if not experiments.STRICT_OPTIONAL:
+            return
+        # Indicate that type should be wrapped in an Optional if arg is initialized to None.
+        optional = isinstance(initializer, NameExpr) and initializer.name == 'None'
+        if isinstance(type, UnboundType):
+            type.optional = optional
+
     def parse_parameter_annotation(self) -> Node:
         if self.current_str() == ':':
             self.skip()
@@ -852,7 +862,7 @@ class Parser:
             if allow_type:
                 cur = self.current()
                 if type is None and isinstance(cur, StrLit):
-                    ds = docstring.parse_docstring(cast(StrLit, cur).parsed())
+                    ds = docstring.parse_docstring(cur.parsed())
                     if ds and False:  # TODO: Enable when this is working.
                         try:
                             type = parse_str_as_signature(ds.as_type_str(), cur.line)
@@ -879,15 +889,13 @@ class Parser:
 
     def try_combine_overloads(self, s: Node, stmt: List[Node]) -> bool:
         if isinstance(s, Decorator) and stmt:
-            fdef = cast(Decorator, s)
+            fdef = s
             n = fdef.func.name()
-            if (isinstance(stmt[-1], Decorator) and
-                    (cast(Decorator, stmt[-1])).func.name() == n):
-                stmt[-1] = OverloadedFuncDef([cast(Decorator, stmt[-1]), fdef])
+            if isinstance(stmt[-1], Decorator) and stmt[-1].func.name() == n:
+                stmt[-1] = OverloadedFuncDef([stmt[-1], fdef])
                 return True
-            elif (isinstance(stmt[-1], OverloadedFuncDef) and
-                    (cast(OverloadedFuncDef, stmt[-1])).name() == n):
-                (cast(OverloadedFuncDef, stmt[-1])).items.append(fdef)
+            elif isinstance(stmt[-1], OverloadedFuncDef) and stmt[-1].name() == n:
+                stmt[-1].items.append(fdef)
                 return True
         return False
 
@@ -1435,7 +1443,7 @@ class Parser:
             items[0] = self.parse_generator_expr(items[0])
         self.expect(']')
         if len(items) == 1 and isinstance(items[0], GeneratorExpr):
-            return ListComprehension(cast(GeneratorExpr, items[0]))
+            return ListComprehension(items[0])
         else:
             expr = ListExpr(items)
             return expr
@@ -1520,7 +1528,7 @@ class Parser:
         expr = SetExpr(items)
         return expr
 
-    def parse_set_comprehension(self, expr: Node):
+    def parse_set_comprehension(self, expr: Node) -> SetComprehension:
         gen = self.parse_generator_expr(expr)
         self.expect('}')
         set_comp = SetComprehension(gen)
@@ -1683,7 +1691,7 @@ class Parser:
         self.expect('.')
         name = self.expect_type(Name)
         if (isinstance(expr, CallExpr) and isinstance(expr.callee, NameExpr)
-                and cast(NameExpr, expr.callee).name == 'super'):
+                and expr.callee.name == 'super'):
             # super() expression
             node = SuperExpr(name.string)  # type: Node
         else:
@@ -1992,7 +2000,7 @@ if __name__ == '__main__':
     # Parse a file and dump the AST (or display errors).
     import sys
 
-    def usage():
+    def usage() -> None:
         print('Usage: parse.py [--py2] [--quiet] FILE [...]', file=sys.stderr)
         sys.exit(2)
 
@@ -2014,7 +2022,9 @@ if __name__ == '__main__':
         s = open(fnam, 'rb').read()
         errors = Errors()
         try:
-            tree = parse(s, fnam, pyversion=pyversion)
+            options = Options()
+            options.python_version = pyversion
+            tree = parse(s, fnam, None, options=options)
             if not quiet:
                 print(tree)
         except CompileError as e:
