@@ -5,7 +5,7 @@ from typing import Callable, cast, List, Tuple
 from mypy.types import (
     Type, UnboundType, TypeVarType, TupleType, UnionType, Instance, AnyType, CallableType,
     Void, NoneTyp, DeletedType, TypeList, TypeVarDef, TypeVisitor, StarType, PartialType,
-    EllipsisType, TypeType
+    EllipsisType, UninhabitedType, TypeType
 )
 from mypy.nodes import (
     BOUND_TVAR, TYPE_ALIAS, UNBOUND_IMPORTED,
@@ -16,6 +16,7 @@ from mypy.sametypes import is_same_type
 from mypy.exprtotype import expr_to_unanalyzed_type, TypeTranslationError
 from mypy.subtypes import satisfies_upper_bound
 from mypy import nodes
+from mypy import experiments
 
 
 type_constructors = {'typing.Tuple', 'typing.Union', 'typing.Callable', 'typing.Type'}
@@ -73,6 +74,11 @@ class TypeAnalyser(TypeVisitor[Type]):
         self.fail = fail_func
 
     def visit_unbound_type(self, t: UnboundType) -> Type:
+        if t.optional:
+            t.optional = False
+            # We don't need to worry about double-wrapping Optionals or
+            # wrapping Anys: Union simplification will take care of that.
+            return UnionType.make_simplified_union([self.visit_unbound_type(t), NoneTyp()])
         sym = self.lookup(t.name, t)
         if sym is not None:
             if sym.node is None:
@@ -85,21 +91,24 @@ class TypeAnalyser(TypeVisitor[Type]):
                 if len(t.args) > 0:
                     self.fail('Type variable "{}" used with arguments'.format(
                         t.name), t)
-                tvar_expr = cast(TypeVarExpr, sym.node)
-                return TypeVarType(t.name, sym.tvar_id, tvar_expr.values,
-                                   tvar_expr.upper_bound,
-                                   tvar_expr.variance,
-                                   t.line)
+                assert sym.tvar_def is not None
+                return TypeVarType(sym.tvar_def, t.line)
             elif fullname == 'builtins.None':
-                return Void()
+                if experiments.STRICT_OPTIONAL:
+                    if t.is_ret_type:
+                        return Void()
+                    else:
+                        return NoneTyp()
+                else:
+                    return Void()
             elif fullname == 'typing.Any':
                 return AnyType()
             elif fullname == 'typing.Tuple':
                 if len(t.args) == 2 and isinstance(t.args[1], EllipsisType):
                     # Tuple[T, ...] (uniform, variable-length tuple)
                     node = self.lookup_fqn_func('builtins.tuple')
-                    info = cast(TypeInfo, node.node)
-                    return Instance(info, [t.args[0].accept(self)], t.line)
+                    tuple_info = cast(TypeInfo, node.node)
+                    return Instance(tuple_info, [t.args[0].accept(self)], t.line)
                 return self.tuple_type(self.anal_array(t.args))
             elif fullname == 'typing.Union':
                 items = self.anal_array(t.args)
@@ -110,8 +119,11 @@ class TypeAnalyser(TypeVisitor[Type]):
                     self.fail('Optional[...] must have exactly one type argument', t)
                     return AnyType()
                 items = self.anal_array(t.args)
-                # Currently Optional[t] is just an alias for t.
-                return items[0]
+                if experiments.STRICT_OPTIONAL:
+                    return UnionType.make_simplified_union([items[0], NoneTyp()])
+                else:
+                    # Without strict Optional checking Optional[t] is just an alias for t.
+                    return items[0]
             elif fullname == 'typing.Callable':
                 return self.analyze_callable_type(t)
             elif fullname == 'typing.Type':
@@ -137,7 +149,7 @@ class TypeAnalyser(TypeVisitor[Type]):
                     return AnyType()
                 self.fail('Invalid type "{}"'.format(name), t)
                 return t
-            info = cast(TypeInfo, sym.node)
+            info = sym.node  # type: TypeInfo
             if len(t.args) > 0 and info.fullname() == 'builtins.tuple':
                 return TupleType(self.anal_array(t.args),
                                  Instance(info, [AnyType()], t.line),
@@ -172,6 +184,9 @@ class TypeAnalyser(TypeVisitor[Type]):
     def visit_none_type(self, t: NoneTyp) -> Type:
         return t
 
+    def visit_uninhabited_type(self, t: UninhabitedType) -> Type:
+        return t
+
     def visit_deleted_type(self, t: DeletedType) -> Type:
         return t
 
@@ -262,7 +277,7 @@ class TypeAnalyser(TypeVisitor[Type]):
     def anal_var_defs(self, var_defs: List[TypeVarDef]) -> List[TypeVarDef]:
         a = []  # type: List[TypeVarDef]
         for vd in var_defs:
-            a.append(TypeVarDef(vd.name, vd.id, self.anal_array(vd.values),
+            a.append(TypeVarDef(vd.name, vd.id.raw_id, self.anal_array(vd.values),
                                 vd.upper_bound.accept(self),
                                 vd.variance,
                                 vd.line))
@@ -384,6 +399,9 @@ class TypeAnalyserPass3(TypeVisitor[None]):
     def visit_none_type(self, t: NoneTyp) -> None:
         pass
 
+    def visit_uninhabited_type(self, t: UninhabitedType) -> None:
+        pass
+
     def visit_deleted_type(self, t: DeletedType) -> None:
         pass
 
