@@ -1,11 +1,15 @@
 """Classes for representing mypy types."""
 
 from abc import abstractmethod
-from typing import Any, TypeVar, Dict, List, Tuple, cast, Generic, Set, Sequence, Optional
+from typing import (
+    Any, TypeVar, Dict, List, Tuple, cast, Generic, Set, Sequence, Optional, Union
+)
 
 import mypy.nodes
 from mypy.nodes import INVARIANT, SymbolNode
 
+from mypy import experiments
+
 
 T = TypeVar('T')
 
@@ -43,25 +47,84 @@ class Type(mypy.nodes.Context):
         raise NotImplementedError('unexpected .class {}'.format(classname))
 
 
+class TypeVarId:
+    # A type variable is uniquely identified by its raw id and meta level.
+
+    # For plain variables (type parameters of generic classes and
+    # functions) raw ids are allocated by semantic analysis, using
+    # positive ids 1, 2, ... for generic class parameters and negative
+    # ids -1, ... for generic function type arguments. This convention
+    # is only used to keep type variable ids distinct when allocating
+    # them; the type checker makes no distinction between class and
+    # function type variables.
+
+    # Metavariables are allocated unique ids starting from 1.
+    raw_id = 0  # type: int
+
+    # Level of the variable in type inference. Currently either 0 for
+    # declared types, or 1 for type inference metavariables.
+    meta_level = 0  # type: int
+
+    # Class variable used for allocating fresh ids for metavariables.
+    next_raw_id = 1  # type: int
+
+    def __init__(self, raw_id: int, meta_level: int = 0) -> None:
+        self.raw_id = raw_id
+        self.meta_level = meta_level
+
+    @staticmethod
+    def new(meta_level: int) -> 'TypeVarId':
+        raw_id = TypeVarId.next_raw_id
+        TypeVarId.next_raw_id += 1
+        return TypeVarId(raw_id, meta_level)
+
+    def __repr__(self) -> str:
+        return self.raw_id.__repr__()
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, TypeVarId):
+            return (self.raw_id == other.raw_id and
+                    self.meta_level == other.meta_level)
+        else:
+            return False
+
+    def __ne__(self, other: object) -> bool:
+        return not (self == other)
+
+    def __hash__(self) -> int:
+        return hash((self.raw_id, self.meta_level))
+
+    def is_meta_var(self) -> bool:
+        return self.meta_level > 0
+
+
 class TypeVarDef(mypy.nodes.Context):
     """Definition of a single type variable."""
 
     name = ''
-    id = 0
-    values = None  # type: Optional[List[Type]]
+    id = None  # type: TypeVarId
+    values = None  # type: List[Type]  # Value restriction, empty list if no restriction
     upper_bound = None  # type: Type
     variance = INVARIANT  # type: int
     line = 0
 
-    def __init__(self, name: str, id: int, values: Optional[List[Type]],
+    def __init__(self, name: str, id: Union[TypeVarId, int], values: Optional[List[Type]],
                  upper_bound: Type, variance: int = INVARIANT, line: int = -1) -> None:
         self.name = name
+        if isinstance(id, int):
+            id = TypeVarId(id)
         self.id = id
         self.values = values
         self.upper_bound = upper_bound
         self.variance = variance
         self.line = line
 
+    @staticmethod
+    def new_unification_variable(old: 'TypeVarDef') -> 'TypeVarDef':
+        new_id = TypeVarId.new(meta_level=1)
+        return TypeVarDef(old.name, new_id, old.values,
+                          old.upper_bound, old.variance, old.line)
+
     def get_line(self) -> int:
         return self.line
 
@@ -74,9 +137,10 @@ class TypeVarDef(mypy.nodes.Context):
             return self.name
 
     def serialize(self) -> JsonDict:
+        assert not self.id.is_meta_var()
         return {'.class': 'TypeVarDef',
                 'name': self.name,
-                'id': self.id,
+                'id': self.id.raw_id,
                 'values': None if self.values is None else [v.serialize() for v in self.values],
                 'upper_bound': self.upper_bound.serialize(),
                 'variance': self.variance,
@@ -99,12 +163,23 @@ class UnboundType(Type):
 
     name = ''
     args = None  # type: List[Type]
+    # should this type be wrapped in an Optional?
+    optional = False
+    # is this type a return type?
+    is_ret_type = False
 
-    def __init__(self, name: str, args: List[Type] = None, line: int = -1) -> None:
+    def __init__(self,
+                 name: str,
+                 args: List[Type] = None,
+                 line: int = -1,
+                 optional: bool = False,
+                 is_ret_type: bool = False) -> None:
         if not args:
             args = []
         self.name = name
         self.args = args
+        self.optional = optional
+        self.is_ret_type = is_ret_type
         super().__init__(line)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
@@ -161,7 +236,7 @@ class TypeList(Type):
 class AnyType(Type):
     """The type 'Any'."""
 
-    def __init__(self, implicit=False, line: int = -1) -> None:
+    def __init__(self, implicit: bool = False, line: int = -1) -> None:
         super().__init__(line)
         self.implicit = implicit
 
@@ -205,17 +280,51 @@ class Void(Type):
         return Void()
 
 
+class UninhabitedType(Type):
+    """This type has no members.
+
+    This type is almost the bottom type, except it is not a subtype of Void.
+    With strict Optional checking, it is the only common subtype between all
+    other types, which allows `meet` to be well defined.  Without strict
+    Optional checking, NoneTyp fills this role.
+
+    In general, for any type T that isn't Void:
+        join(UninhabitedType, T) = T
+        meet(UninhabitedType, T) = UninhabitedType
+        is_subtype(UninhabitedType, T) = True
+    """
+
+    def __init__(self, line: int = -1) -> None:
+        super().__init__(line)
+
+    def accept(self, visitor: 'TypeVisitor[T]') -> T:
+        return visitor.visit_uninhabited_type(self)
+
+    def serialize(self) -> JsonDict:
+        return {'.class': 'UninhabitedType'}
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'UninhabitedType':
+        assert data['.class'] == 'UninhabitedType'
+        return UninhabitedType()
+
+
 class NoneTyp(Type):
     """The type of 'None'.
 
-    This is only used internally during type inference.  Programs
-    cannot declare a variable of this type, and the type checker
-    refuses to infer this type for a variable. However, subexpressions
-    often have this type. Note that this is not used as the result
-    type when calling a function with a void type, even though
-    semantically such a function returns a None value; the void type
-    is used instead so that we can report an error if the caller tries
-    to do anything with the return value.
+    Without strict Optional checking:
+        This is only used internally during type inference.  Programs
+        cannot declare a variable of this type, and the type checker
+        refuses to infer this type for a variable. However, subexpressions
+        often have this type. Note that this is not used as the result
+        type when calling a function with a void type, even though
+        semantically such a function returns a None value; the void type
+        is used instead so that we can report an error if the caller tries
+        to do anything with the return value.
+
+    With strict Optional checking:
+        This type can be written by users as 'None', except as the return value
+        of a function, where 'None' means Void.
     """
 
     def __init__(self, line: int = -1) -> None:
@@ -321,19 +430,18 @@ class TypeVarType(Type):
     """
 
     name = ''  # Name of the type variable (for messages and debugging)
-    id = 0     # 1, 2, ... for type-related, -1, ... for function-related
+    id = None  # type: TypeVarId
     values = None  # type: List[Type]  # Value restriction, empty list if no restriction
     upper_bound = None  # type: Type   # Upper bound for values
     # See comments in TypeVarDef for more about variance.
     variance = INVARIANT  # type: int
 
-    def __init__(self, name: str, id: int, values: List[Type], upper_bound: Type,
-                 variance: int = INVARIANT, line: int = -1) -> None:
-        self.name = name
-        self.id = id
-        self.values = values
-        self.upper_bound = upper_bound
-        self.variance = variance
+    def __init__(self, binder: TypeVarDef, line: int = -1) -> None:
+        self.name = binder.name
+        self.id = binder.id
+        self.values = binder.values
+        self.upper_bound = binder.upper_bound
+        self.variance = binder.variance
         super().__init__(line)
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
@@ -346,9 +454,10 @@ class TypeVarType(Type):
             return self.upper_bound
 
     def serialize(self) -> JsonDict:
+        assert not self.id.is_meta_var()
         return {'.class': 'TypeVarType',
                 'name': self.name,
-                'id': self.id,
+                'id': self.id.raw_id,
                 'values': [v.serialize() for v in self.values],
                 'upper_bound': self.upper_bound.serialize(),
                 'variance': self.variance,
@@ -357,11 +466,12 @@ class TypeVarType(Type):
     @classmethod
     def deserialize(cls, data: JsonDict) -> 'TypeVarType':
         assert data['.class'] == 'TypeVarType'
-        return TypeVarType(data['name'],
+        tvdef = TypeVarDef(data['name'],
                            data['id'],
                            [Type.deserialize(v) for v in data['values']],
                            Type.deserialize(data['upper_bound']),
                            data['variance'])
+        return TypeVarType(tvdef)
 
 
 class FunctionLike(Type):
@@ -428,9 +538,9 @@ class CallableType(FunctionLike):
                  variables: List[TypeVarDef] = None,
                  line: int = -1,
                  is_ellipsis_args: bool = False,
-                 implicit=False,
-                 is_classmethod_class=False,
-                 special_sig=None,
+                 implicit: bool = False,
+                 is_classmethod_class: bool = False,
+                 special_sig: Optional[str] = None,
                  ) -> None:
         if variables is None:
             variables = []
@@ -514,8 +624,8 @@ class CallableType(FunctionLike):
     def is_generic(self) -> bool:
         return bool(self.variables)
 
-    def type_var_ids(self) -> List[int]:
-        a = []  # type: List[int]
+    def type_var_ids(self) -> List[TypeVarId]:
+        a = []  # type: List[TypeVarId]
         for tv in self.variables:
             a.append(tv.id)
         return a
@@ -683,7 +793,10 @@ class UnionType(Type):
         elif len(items) == 1:
             return items[0]
         else:
-            return Void()
+            if experiments.STRICT_OPTIONAL:
+                return UninhabitedType()
+            else:
+                return Void()
 
     @staticmethod
     def make_simplified_union(items: List[Type], line: int = -1) -> Type:
@@ -754,10 +867,15 @@ class PartialType(Type):
     # None for the 'None' partial type; otherwise a generic class
     type = None  # type: Optional[mypy.nodes.TypeInfo]
     var = None  # type: mypy.nodes.Var
+    inner_types = None  # type: List[Type]
 
-    def __init__(self, type: Optional['mypy.nodes.TypeInfo'], var: 'mypy.nodes.Var') -> None:
+    def __init__(self,
+                 type: Optional['mypy.nodes.TypeInfo'],
+                 var: 'mypy.nodes.Var',
+                 inner_types: List[Type]) -> None:
         self.type = type
         self.var = var
+        self.inner_types = inner_types
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_partial_type(self)
@@ -825,7 +943,8 @@ class TypeType(Type):
     def serialize(self) -> JsonDict:
         return {'.class': 'TypeType', 'item': self.item.serialize()}
 
-    def deserialize(self, data: JsonDict) -> 'TypeType':
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'TypeType':
         assert data['.class'] == 'TypeType'
         return TypeType(Type.deserialize(data['item']))
 
@@ -869,6 +988,10 @@ class TypeVisitor(Generic[T]):
     def visit_none_type(self, t: NoneTyp) -> T:
         pass
 
+    @abstractmethod
+    def visit_uninhabited_type(self, t: UninhabitedType) -> T:
+        pass
+
     def visit_erased_type(self, t: ErasedType) -> T:
         raise self._notimplemented_helper('erased_type')
 
@@ -939,6 +1062,9 @@ class TypeTranslator(TypeVisitor[Type]):
     def visit_none_type(self, t: NoneTyp) -> Type:
         return t
 
+    def visit_uninhabited_type(self, t: UninhabitedType) -> Type:
+        return t
+
     def visit_erased_type(self, t: ErasedType) -> Type:
         return t
 
@@ -1006,38 +1132,41 @@ class TypeStrVisitor(TypeVisitor[str]):
      - Represent the NoneTyp type as None.
     """
 
-    def visit_unbound_type(self, t):
+    def visit_unbound_type(self, t: UnboundType)-> str:
         s = t.name + '?'
         if t.args != []:
             s += '[{}]'.format(self.list_str(t.args))
         return s
 
-    def visit_type_list(self, t):
+    def visit_type_list(self, t: TypeList) -> str:
         return '<TypeList {}>'.format(self.list_str(t.items))
 
-    def visit_error_type(self, t):
+    def visit_error_type(self, t: ErrorType) -> str:
         return '<ERROR>'
 
-    def visit_any(self, t):
+    def visit_any(self, t: AnyType) -> str:
         return 'Any'
 
-    def visit_void(self, t):
+    def visit_void(self, t: Void) -> str:
         return 'void'
 
-    def visit_none_type(self, t):
-        # Include quotes to make this distinct from the None value.
-        return "'None'"
+    def visit_none_type(self, t: NoneTyp) -> str:
+        # Fully qualify to make this distinct from the None value.
+        return "builtins.None"
 
-    def visit_erased_type(self, t):
+    def visit_uninhabited_type(self, t: UninhabitedType) -> str:
+        return "<uninhabited>"
+
+    def visit_erased_type(self, t: ErasedType) -> str:
         return "<Erased>"
 
-    def visit_deleted_type(self, t):
+    def visit_deleted_type(self, t: DeletedType) -> str:
         if t.source is None:
             return "<Deleted>"
         else:
             return "<Deleted '{}'>".format(t.source)
 
-    def visit_instance(self, t):
+    def visit_instance(self, t: Instance) -> str:
         s = t.type.fullname() if t.type is not None else '<?>'
         if t.erased:
             s += '*'
@@ -1045,7 +1174,7 @@ class TypeStrVisitor(TypeVisitor[str]):
             s += '[{}]'.format(self.list_str(t.args))
         return s
 
-    def visit_type_var(self, t):
+    def visit_type_var(self, t: TypeVarType) -> str:
         if t.name is None:
             # Anonymous type variable type (only numeric id).
             return '`{}'.format(t.id)
@@ -1053,7 +1182,7 @@ class TypeStrVisitor(TypeVisitor[str]):
             # Named type variable type.
             return '{}`{}'.format(t.name, t.id)
 
-    def visit_callable_type(self, t):
+    def visit_callable_type(self, t: CallableType) -> str:
         s = ''
         bare_asterisk = False
         for i in range(len(t.arg_types)):
@@ -1082,13 +1211,13 @@ class TypeStrVisitor(TypeVisitor[str]):
 
         return 'def {}'.format(s)
 
-    def visit_overloaded(self, t):
+    def visit_overloaded(self, t: Overloaded) -> str:
         a = []
         for i in t.items():
             a.append(i.accept(self))
         return 'Overload({})'.format(', '.join(a))
 
-    def visit_tuple_type(self, t):
+    def visit_tuple_type(self, t: TupleType) -> str:
         s = self.list_str(t.items)
         if t.fallback and t.fallback.type:
             fallback_name = t.fallback.type.fullname()
@@ -1096,11 +1225,11 @@ class TypeStrVisitor(TypeVisitor[str]):
                 return 'Tuple[{}, fallback={}]'.format(s, t.fallback.accept(self))
         return 'Tuple[{}]'.format(s)
 
-    def visit_star_type(self, t):
+    def visit_star_type(self, t: StarType) -> str:
         s = t.type.accept(self)
         return '*{}'.format(s)
 
-    def visit_union_type(self, t):
+    def visit_union_type(self, t: UnionType) -> str:
         s = self.list_str(t.items)
         return 'Union[{}]'.format(s)
 
@@ -1111,13 +1240,13 @@ class TypeStrVisitor(TypeVisitor[str]):
             return '<partial {}[{}]>'.format(t.type.name(),
                                              ', '.join(['?'] * len(t.type.type_vars)))
 
-    def visit_ellipsis_type(self, t):
+    def visit_ellipsis_type(self, t: EllipsisType) -> str:
         return '...'
 
-    def visit_type_type(self, t):
+    def visit_type_type(self, t: TypeType) -> str:
         return 'Type[{}]'.format(t.item.accept(self))
 
-    def list_str(self, a):
+    def list_str(self, a: List[Type]) -> str:
         """Convert items of an array to strings (pretty-print types)
         and join the results with commas.
         """
@@ -1172,6 +1301,9 @@ class TypeQuery(TypeVisitor[bool]):
     def visit_void(self, t: Void) -> bool:
         return self.default
 
+    def visit_uninhabited_type(self, t: UninhabitedType) -> bool:
+        return self.default
+
     def visit_none_type(self, t: NoneTyp) -> bool:
         return self.default
 
